
class AppError(Exception):
    pass


class InputError(Exception):
    """Base class for other input exceptions"""
    pass


class AddContactInputError(InputError):
    def __str__(self):
        return "AddContactInputError: 'add-contact' command expects two arguments 'name' and 'phone'"


class AddBirthdatInputError(InputError):
    def __str__(self):
        return "AddBirthdatInputError: 'add-birthday' command expects two arguments 'name' and 'birthday'."


class ChangeInputError(InputError):
    def __str__(self):
        return "ChangeInputError: 'change-phone' command expects two arguments 'name' and 'phone'"

class ChangeEmailInputError(InputError):
    def __str__(self):
        return "ChangeInputError: 'change-email' command expects two arguments 'name' and  'email'."

class FindContactsInputError(InputError):
    def __str__(self):
        return "FindContactsInputError: 'find-contacts' command expects one arguments 'name'."


class AddAddressInputError(InputError):
    def __str__(self):
        return "AddBirthdatInputError: 'add-address' command expects two arguments 'name' and 'address'."

class PhoneInputError(InputError):
    def __str__(self):
        return "PhoneInputError: 'show-phone' command expects one argument 'name'."


class GetBirthdayInputError(InputError):
    def __str__(self):
        return "GetBirthdayInputError: 'show-birthday' command expects one argument username"


class InaccurateBirthdayFormat(InputError):
    def __str__(self):
        return "InaccurateBirthdayFormat: Birthday must be in format DD.MM.YYYY."


class InaccuratePhoneFormat(InputError):
    def __str__(self):
        return "InaccuratePhoneFormat: Phone number must contain exactly 10 digits."

class InaccurateEmailFormat(InputError):
    def __str__(self):
        return "InaccurateEmailFormat: The email address provided does not adhere to the standard email format."


class NonExistingContact(InputError):
    def __str__(self):
        return "NonExistingContact: Check contact name, no such contact in address book"


def input_error(func):
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except InputError as e:
            return str(e)

    return inner



DATE_FORMAT = "%d.%m.%Y"
ADDRESSBOOK_FILE = "addressbook.json"
NOTES_FILE = "notes.json"
WEEKDAYS = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday", "Next Monday"]
MONDAY_INDEX = 0
NEXT_MONDAY_INDEX = -1
PROMPT_STYLE = {
    # User input (default text).
    '': '#ffffff',
    # Prompt, approximation of Fore.LIGHTGREEN_EX
    'input_request': '#32c986',
}
PROMPT_MESSAGE = [
    ('class:input_request', 'Enter a command:'),
]



import re
from datetime import date, datetime
from typing import List, Optional

from contacts24.errors import InaccurateBirthdayFormat, InaccuratePhoneFormat, InaccurateEmailFormat
from contacts24.config import DATE_FORMAT


class Field:
    def __init__(self, value: str):
        self.value = value

    def __str__(self) -> str:
        return str(self.value)


class Name(Field):
    pass

class Address(Field):
    pass


class Email(Field):
    def __init__(self, value: str):
        if not re.match(r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$", value):
            raise InaccurateEmailFormat()
        super().__init__(value)

class Phone(Field):
    def __init__(self, value: str):
        if not re.match(r"^\d{10}$", value):
            raise InaccuratePhoneFormat()
        super().__init__(value)


class Birthday(Field):
    def __init__(self, value: str):
        if not re.match(r"^\d{2}\.\d{2}\.\d{4}$", value):
            raise InaccurateBirthdayFormat()
        super().__init__(value)

    def get_birthday_datetime(self) -> date:
        return datetime.strptime(self.value, DATE_FORMAT).date()


class Record:
    def __init__(self, name: str):
        self.name = Name(name)
        self.phones: List[Phone] = []
        self.emails: List[Email] = []
        self.birthday: Optional[Birthday] = None
        self.address: Optional[Address] = None

    def add_birthday(self, birthday: str) -> None:
        self.birthday = Birthday(birthday)

    def add_address(self, address: str) -> None:
        self.address = Address(address)

    def add_phone(self, phone: str) -> None:
        new_phone = Phone(phone)
        if new_phone:
            self.phones.append(new_phone)

    def add_email(self, email: str) -> None:
        new_email = Email(email)
        if new_email:
            self.emails.append(new_email)

    def delete_phone(self, phone: str) -> None:
        self.phones = [p for p in self.phones if p.value != phone]

    def delete_email(self, email: str) -> None:
        self.emails = [p for p in self.emails if p.value != email]

    def edit_phone(self, old_phone: str, new_phone: str) -> None:
        for index, p in enumerate(self.phones):
            if p.value == old_phone:
                self.phones[index] = Phone(new_phone)
                break

    def edit_email(self, old_email: str, new_email: str) -> None:
        for index, p in enumerate(self.emails):
            if p.value == old_email:
                self.emails[index] = Email(new_email)
                break

    def find_phone(self, phone: str) -> Optional[str]:
        for p in self.phones:
            if p.value == phone:
                return p.value
        return None

    def find_email(self, email: str) -> Optional[str]:
        for p in self.emails:
            if p.value == email:
                return p.value
        return None

    def __str__(self) -> str:
        return (
            f"Contact Information:\n"
            f"---------------------\n"
            f"Name: {self.name}\n"
            f"Birthday: {self.birthday if self.birthday else 'N/A'}\n"
            f"Address: {self.address if self.address else 'N/A'}\n"
            f"Phones: {', '.join(p.value for p in self.phones) if self.phones else 'N/A'}\n"
            f"Emails: {', '.join(p.value for p in self.emails) if self.emails else 'N/A'}\n"
        )



from contacts24.models.record import Record
from contacts24.models.note import Note

#region Record serialization

def record_serialization(record: Record) -> dict:
    """Record to dictionary

    Args:
        record (Record): input

    Returns:
        dict: dictionary based on input record
    """
    return {
            "name": record.name.value,
            "phones": [phone.value for phone in record.phones],
            "emails": [email.value for email in record.emails],
            "birthday": record.birthday.value if record.birthday else None,
            "address": record.address.value if record.address else None,
        }

def record_deserialization(data: dict) -> Record:
    """Deserialization Record from json

    Args:
        data (dict): Record in json

    Returns:
        Record: instance of class Record
    """
    record = Record(data["name"])
    
    for phone in data["phones"]:
        record.add_phone(phone)
    for email in data["emails"]:
        record.add_email(email)
    if data["birthday"]:
        record.add_birthday(data["birthday"])
    if data["address"]:
        record.add_address(data["address"])
    
    return record

#endregion

#region Note serialization

def note_serialization(note: Note) -> dict:
    """Note to dictionary

    Args:
        note (Note): input

    Returns:
        dict: dictionary based on input record
    """
    return {
        "id": note.name.value,
        "text": note.text.value,
    }

def note_deserialization(data: dict) -> Note:
    """Deserialization Note from json

    Args:
        data (dict): Note in json

    Returns:
        Note: instance of class Note
    """
    note = Note()
    
    note.add_text(data["text"], data["id"])
    
    return note

#endregion



from collections import UserDict

from contacts24.models.note import Note

# Temporary contract
class Notes(UserDict):
    def add_note(self, note: Note) -> None:
        # To complete by Sergii Shulga
        pass

def get_notes():
    pass


def create_note():
    pass


def update_note():
    pass


def delete_note():
    pass



"""CLI assistant functions"""

from typing import Optional

from contacts24.config import ADDRESSBOOK_FILE
from .models.address_book import AddressBook
from .models.record import Record
from .errors import (
    AddBirthdatInputError,
    AddContactInputError,
    ChangeInputError,
    ChangeEmailInputError,
    FindContactsInputError,
    AddAddressInputError,
    GetBirthdayInputError,
    NonExistingContact,
    PhoneInputError,
    input_error,
)

Contacts = dict[str, str]
CommandArguments = list[str]


def parse_input(user_input: str) -> tuple[str, Optional[CommandArguments]]:
    if not user_input.strip():
        return "", None

    cmd, *args = user_input.split(" ")
    cmd = cmd.strip().lower()

    if len(args) == 0:
        return cmd, None

    return cmd, args

def load_contacts_book() -> AddressBook:
    try:
        contacts = AddressBook.load_from_file(ADDRESSBOOK_FILE)
    except FileNotFoundError:
        print(f"File {ADDRESSBOOK_FILE} not found. Initializing an empty AddressBook.")
        contacts = AddressBook()
    return contacts


@input_error
def add_contact(args: CommandArguments, contacts: AddressBook) -> str:
    if args is None or len(args) < 2:
        raise AddContactInputError()

    name, phone= args[:2]
    new_contact = Record(name)
    new_contact.add_phone(phone)
    contacts.add_record(new_contact)
    return f"Contact {name} added."


@input_error
def add_birthday(args: CommandArguments, contacts: AddressBook) -> str:
    if args is None or len(args) < 2:
        raise AddBirthdatInputError()

    name, date = args[:2]
    contact = contacts.find(name)
    if not contact:
        raise NonExistingContact()

    contact.add_birthday(date)
    return f"Birthday added for {name} updated."


@input_error
def change_contact(args: CommandArguments, contacts: AddressBook) -> str:
    if args is None or len(args) < 2:
        raise ChangeInputError()

    name, phone = args[:2]
    contact = contacts.find(name)
    if not contact:
        raise NonExistingContact()

    contact.edit_phone(contact.phones[0].value, phone)
    return f"Phone number, email and address for {name} updated."


@input_error
def get_contact_phone(args: CommandArguments, contacts: AddressBook) -> str:
    if args is None:
        raise PhoneInputError()

    name = args[0]
    contact = contacts.find(name)
    if not contact:
        raise NonExistingContact()

    return contact.phones[0].value


@input_error
def get_contact_birthday(args: CommandArguments, contacts: AddressBook) -> str:
    if args is None:
        raise GetBirthdayInputError()

    name = args[0]
    contact = contacts.find(name)
    if not contact:
        raise NonExistingContact()

    return contact.birthday


@input_error
def get_upcoming_birthdays(args: CommandArguments, contacts: AddressBook) -> str:
    return contacts.get_birthdays_per_week()


def get_all_contacts(args: CommandArguments, contacts: AddressBook) -> str:
    if not contacts:
        return "No contacts stored."
    else:
        return str(contacts)


def find_contacts(args: CommandArguments, contacts: AddressBook) -> str:
    if args is None or len(args) < 1:
        raise FindContactsInputError()

    name = args[0]
    result_book = AddressBook()

    for contact in contacts.data.values():
        if name in str(contact.name):
            result_book.add_record(contact)

    if len(result_book) == 0:
        return "No contacts found."
    elif len(result_book) == 1:
        contact = result_book.find(list(result_book.data.keys())[0])
        return contact
    else:
        return result_book
    

@input_error
def change_email(args: CommandArguments, contacts: AddressBook) -> str:
    if args is None or len(args) < 2:
        raise ChangeEmailInputError()

    name, email = args[:2]
    contact = contacts.find(name)
    if not contact:
        raise NonExistingContact()

    if  len(contact.emails):
        contact.edit_email(contact.emails[0].value, email)
    else:
        contact.add_email(email)

    return f"Email for {name} updated."

@input_error
def add_address(args: CommandArguments, contacts: AddressBook) -> str:
    if args is None or len(args) < 2:
        raise AddAddressInputError()

    name, address = args[:2]
    contact = contacts.find(name)
    if not contact:
        raise NonExistingContact()

    contact.add_address(address)
    return f"Address for {name} updated."



"""Prototype of CLI assistant"""

from functools import partial

from colorama import Fore, init
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.styles import Style
from .cli_commands import Command, get_help

from .config import PROMPT_MESSAGE, PROMPT_STYLE
from .contacts import (
    add_birthday,
    add_contact,
    change_contact,
    change_email,
    add_address,
    get_all_contacts,
    find_contacts,
    get_contact_birthday,
    get_contact_phone,
    get_upcoming_birthdays,
    load_contacts_book,
    parse_input,
)

init(autoreset=True)


contacts = load_contacts_book()

commands = {
    "hello": Command("hello", lambda x: "How can I help you?", is_hidden = True),
    "help": Command("help",  lambda x: "\n\nAvailable commands: \n{0}\n".format(get_help(commands)), is_hidden=True),
    "add-contact": Command("add-contact",  partial(add_contact, contacts=contacts), "Adds a new contact", "add-contact <username> <phone>"),
    "change-phone": Command("change-phone", partial(change_contact, contacts=contacts), "Changes a contact's phone number", "change-phone <username> <phone>"), 
    "change-email": Command("change-email", partial(change_email, contacts=contacts), "Changes a contact's email", "change-email <username> <email>"),
    "change-address": Command("change-address", partial(add_address, contacts=contacts), "Changes a contact's address", "change-email <username> <address>"),
    "show-phone": Command("show-phone", partial(get_contact_phone, contacts=contacts), "Prints a contact's number", "show-phone <username>"), 
    "add-birthday": Command("add-birthday", partial(add_birthday, contacts=contacts),  "Adds a contact's birthday", "add-birthday <username> <date>"), 
    "show-birthday": Command("show-birthday", partial(get_contact_birthday, contacts=contacts),  "Prints a contact's birthday", "show-birthday <username>"), 
    "birthdays": Command("birthdays", partial(get_upcoming_birthdays, contacts=contacts), "Prints upcoming birthdays", "birthdays"), 
    "show-all-contacts": Command("show-all-contacts", partial(get_all_contacts, contacts=contacts), "Prints all contacts", "show-all-contacts"), 
    "find-contacts": Command("find-contact", partial(find_contacts, contacts=contacts), "Prints contacts by name", "find-contacts"),
    "add-note": Command("add-note", lambda x: "In development"),
    "change-note": Command("change-note", lambda x: "In development"), 
    "find-note": Command("find-note", lambda x: "In development"), 
    "show-all-notes": Command("show-all-notes", lambda x: "In development"), 
}


def start():
    print(Fore.LIGHTGREEN_EX + "Welcome to the Contacts24!\n" + Fore.WHITE + "Type 'help' to see available commands.\n")
    command_completer = WordCompleter(list(commands.keys()))
    while True:
        try:
            style = Style.from_dict(PROMPT_STYLE)
            user_input = prompt(PROMPT_MESSAGE, style=style, completer=command_completer)
            command, args = parse_input(user_input)

            if command == "exit":
                print(Fore.LIGHTGREEN_EX + "Good bye!")
                contacts.save_to_file()
                break

            if command in commands:
                print(Fore.WHITE + str(commands[command]()(args)))
            else:
                help_text = get_help(commands)
                print(Fore.RED + f"Unknown command '{command}', please try again.\n{help_text}")
        except Exception as e:
            print(Fore.RED + f"Unexpected error occured {e}")



from collections import UserDict, defaultdict
from datetime import date, datetime
from typing import Optional

from contacts24.models.record import Record
from contacts24.config import (
    ADDRESSBOOK_FILE, WEEKDAYS, MONDAY_INDEX, NEXT_MONDAY_INDEX, DATE_FORMAT
)


class AddressBook(UserDict):
    def add_record(self, record: Record) -> None:
        self.data[record.name.value] = record

    def find(self, name: str) -> Optional[Record]:
        return self.data.get(name)

    def delete(self, name: str) -> None:
        if name in self.data:
            del self.data[name]
            print(f"{name} record is deleted")
        else:
            print(f"No records found by the name {name}")

    def save_to_file(self, filename: str = ADDRESSBOOK_FILE) -> None:
        from contacts24.db import save_address_book
        if filename:
            save_address_book(self, filename)
        else:
            save_address_book(self)

    @staticmethod
    def load_from_file(filename: str = ADDRESSBOOK_FILE) -> "AddressBook":
        from contacts24.db import get_contacts
        if filename:
            return get_contacts(filename)
        else:
            return get_contacts()

    def __str__(self) -> str:
        # Initialize the header and its length
        columns = ['Name', 'Birthday', 'Address', 'Phones', 'Emails']
        col_width = [len(h)+ 3 for h in columns]

        # Gather all data
        data = []
        for name, record in self.data.items():
            phones = ', '.join(str(p.value) for p in record.phones) if record.phones else 'N/A'
            emails = ', '.join(str(e.value) for e in record.emails) if record.emails else 'N/A'
            birthday = str(record.birthday) if record.birthday else 'N/A'
            address = str(record.address) if record.address else 'N/A'
            row = [name, birthday, address, phones, emails]
            
            # Update maximum column width
            col_width = [max(col_width[i], len(cell)) for i, cell in enumerate(row)]

            data.append(row)

        # Define format string for each row
        format_string = ''.join(f'{{:<{w}}} ' for w in col_width)

        # Build table string
        output = format_string.format(*columns) + '\n'
        output += '\n'.join(format_string.format(*row) for row in data)

        return output

    def get_birthdays_per_week(self, relative_date: date = datetime.today().date()) -> str:
        """
        Determine which colleagues have birthdays in the upcoming week.

        Args:
            relative_date (date, optional): The date from which to calculate one week forward. Default is today's date.

        Returns:
            str: A formatted string containing the names of colleagues that have birthdays
                in the upcoming week, sorted by the day of the week.
        """
        next_week_birthday_colleagues = defaultdict(list)

        for colleague in self.data.values():
            name = colleague.name
            if not colleague.birthday:
                continue
            birthday = colleague.birthday.get_birthday_datetime()
            birthday_this_year = self._get_birthday_this_year(birthday, relative_date)

            delta_days = (birthday_this_year - relative_date).days
            if delta_days < 7:
                weekday_to_greet = self._get_greeting_day(birthday_this_year, relative_date)
                next_week_birthday_colleagues[weekday_to_greet].append(
                    f"{name} ({birthday_this_year.strftime(DATE_FORMAT)})"
                )

        relative_date_str = f"{relative_date.strftime(DATE_FORMAT)}, {WEEKDAYS[relative_date.weekday()]}"

        greeting_string = f"---\nColleagues to greet for the next week, as of {relative_date_str}:\n---"
        if next_week_birthday_colleagues:
            for week_day in WEEKDAYS:
                week_day_birthdays = next_week_birthday_colleagues.get(week_day)
                if week_day_birthdays:
                    greeting_string += f"\n{week_day}: {', '.join(week_day_birthdays)}"
        else:
            greeting_string += "\nNo birthdays this week :("

        return greeting_string

    @staticmethod
    def _get_birthday_this_year(birthday: date, relative_date: date) -> date:
        """
        Returns this year's birthday based on a given reference date.

        Args:
            birthday (date): The birthday date.
            relative_date (date): The reference date.

        Returns:
            date: This year's birthday based on the reference date.
        """

        birthday_this_year = birthday.replace(year=relative_date.year)
        if birthday_this_year < relative_date:
            birthday_this_year = birthday_this_year.replace(year=relative_date.year + 1)

        return birthday_this_year

    @staticmethod
    def _get_greeting_day(birthday_this_year: date, relative_date: date) -> str:
        """
        Returns the weekday to greet the colleague.

        Args:
            birthday_this_year (date): The colleague's birthday this year.
            relative_date (date): The reference date.

        Returns:
            str: The weekday to greet the colleague.
        """

        # if birthday is a weekend and relative_date is Monday, so greeting goes to next Monday
        if (birthday_this_year.weekday() >= 5) & (relative_date.weekday() == 0):
            return WEEKDAYS[NEXT_MONDAY_INDEX]
        elif birthday_this_year.weekday() >= 5:
            return WEEKDAYS[MONDAY_INDEX]
        else:
            return WEEKDAYS[birthday_this_year.weekday()]



from typing import Dict 


class Command():
    """Command"""
    def __init__(self, name: str, func, description="", format = "", is_hidden = False):
        """_summary_

        Args:
            name (str): _description_
            func (function|lambda): command function
            description (str, optional): text description of the command. Defaults to "".
            format (str, optional): how to call command. Defaults to "".
            is_hidden (bool, optional): Is command hidden from the help output. Defaults to False.
        """
        self.name = name
        self.func = func
        self.description = description
        self.format = format
        self.is_hidden = is_hidden
        
    def __call__(self):
        return self.func


def get_help(commands: Dict[str, Command]) -> str:
    """Generate help based on the available list of commands

    Returns:
        str: Help table
    """
    message = "\n"
    
    padding_command = len("Command")
    padding_description = len("Description")
    padding_format = len("Format")
    
    for name, command in commands.items():
        if command.is_hidden:
            continue
        padding_command = max([padding_command, len(name)])
        padding_description = max([padding_description, len(command.description)])
        padding_format = max([padding_format, len(command.format)])
    
    border = (("| {0:-^%s} " % padding_command) + ("| {0:-^%s} |" % padding_description) + (" {0:-^%s} |\n" % padding_format)).format("-")
    
    message += border
    message += (("| {0:^%s} " % padding_command) + ("| {1:^%s} |" % padding_description) + (" {2:^%s} |\n" % padding_format)).format("Command", "Description", "Format")
    message += border
    
    for name, command in commands.items():
        if command.is_hidden:
            continue
        message += (("| {0:<%s} " % padding_command) + ("| {1:<%s} |" % padding_description) + (" {2:<%s} |\n" % padding_format)).format(command.name, command.description, command.format)
    message += border
    
    return message



import json

from contacts24.models.address_book import AddressBook
from contacts24.models.notes import Notes
from contacts24.config import ADDRESSBOOK_FILE, NOTES_FILE
from contacts24.errors import AppError
from contacts24.serialization_helper import record_deserialization, record_serialization, note_deserialization, note_serialization


#region AddressBook

def get_contacts(filename: str = ADDRESSBOOK_FILE) -> AddressBook:
    """Get addressbook from json file (default config.ADDRESSBOOK_FILE)

    Raises:
        AppError: Any errors (incorrect json, etc.) wrapped in AppError

    Returns:
        AddressBook: AddressBook from json
    """
    address_book = AddressBook()
    
    try:
        with open(filename, "r") as file:
            records_list = json.load(file)
            
        for record_dict in records_list:
            record = record_deserialization(record_dict)
            address_book.add_record(record)
    except Exception as e:
        raise AppError(e)
    
    return address_book

def save_address_book(address_book: AddressBook, filename: str = ADDRESSBOOK_FILE) -> None:
    """Save address book to json file (from config.ADDRESSBOOK_FILE)

    Args:
        address_book (AddressBook): Address Book to save
        filename (str): file name

    Raises:
        AppError: Any errors (missing file, incorrect json) wrapped in AppError
    """
    try:
        with open(filename, "w") as file:
            records_list = [record_serialization(record) for record in address_book.data.values()]
            json.dump(records_list, file)
    except Exception as e:
        raise AppError(e)

#endregion

#region Notes

def get_notes(filename: str = NOTES_FILE) -> Notes:
    """Get notes from fixed json file (from config.NOTES_FILE)

    Raises:
        AppError: All errors wrapped in AppError

    Returns:
        Notes: List of notes
    """
    notes = Notes()
    
    try:
        with open(filename, "r") as file:
            notes_list = json.load(file)
            for note_dict in notes_list:
                note = note_deserialization(note_dict)
                notes.add_note(note)
    except Exception as e:
        raise AppError(e)
    
    return notes

def save_notes(notes: Notes, filename: str = NOTES_FILE) -> None:
    """Save notes to json file (from config.NOTES_FILE)

    Args:
        notes (Notes): Notes to save

    Raises:
        AppError: Any errors (missing file, incorrect json) wrapped in AppError
    """

    try:
        with open(filename, "w") as file:
            notes_list = [note_serialization(note) for note in notes.data.values()]
            json.dump(notes_list, file)
    except Exception as e:
        raise AppError(e)

#endregion